import xmltodict

from boinc_client.clients.rpc_client import RpcClient
from boinc_client.models.client_update import ClientUpdate
from boinc_client.models.client_version import ClientVersion


def client_version(client: RpcClient) -> dict:
    """Used to get the version of the running core client and send the version of the request's source."""
    rpc_resp = client.make_request("<exchange_versions/>")
    rpc_json = xmltodict.parse(rpc_resp)
    return ClientVersion().load(rpc_json)


def client_update(client: RpcClient) -> dict:
    """Get newer version number, if any, and download url."""
    rpc_resp = client.make_request("<get_newer_version/>")
    update_root = f"<update>{rpc_resp}</update>"
    rpc_json = xmltodict.parse(update_root)
    return ClientUpdate().load(rpc_json)



from boinc_client.clients.rpc_client import RpcClient

from .messages import message_count, messages, public_notices
from .projects import (
    all_projects,
    attach_project,
    detach_project,
    poll_attach_project,
    update_project,
)
from .results import old_results, results
from .stats import daily_network_transfers, project_stats
from .status import (
    cc_status,
    client_state,
    disk_stats,
    file_transfers,
    host_info,
    project_status,
    screensaver_tasks,
    simple_gui_info,
)
from .versions import client_update, client_version


class Boinc:
    rpc_client: RpcClient

    def __init__(self, rpc_client: RpcClient):
        self.rpc_client = rpc_client

    # Messages
    def get_messages(self, start: int = 0) -> dict:
        return messages(client=self.rpc_client, start=start)

    def get_message_count(self) -> dict:
        return message_count(client=self.rpc_client)

    def get_public_notices(self, start: int = 0) -> dict:
        return public_notices(client=self.rpc_client, start=start)

    # Projects
    def get_all_projects(self) -> dict:
        return all_projects(client=self.rpc_client)

    # Results
    def get_results(self, active_only: bool = False) -> dict:
        return results(client=self.rpc_client, active_only=active_only)

    def get_old_results(self) -> dict:
        return old_results(client=self.rpc_client)

    # Stats
    def get_network_stats(self) -> dict:
        return daily_network_transfers(client=self.rpc_client)

    def get_project_stats(self) -> dict:
        return project_stats(client=self.rpc_client)

    # Status
    def get_client_state(self) -> dict:
        return client_state(client=self.rpc_client)

    def get_project_status(self) -> dict:
        return project_status(client=self.rpc_client)

    def get_cc_status(self) -> dict:
        return cc_status(client=self.rpc_client)

    def get_disk_stats(self) -> dict:
        return disk_stats(client=self.rpc_client)

    def get_file_transfers(self) -> dict:
        return file_transfers(client=self.rpc_client)

    def get_host_info(self) -> dict:
        return host_info(client=self.rpc_client)

    def get_simple_gui_info(self) -> dict:
        return simple_gui_info(client=self.rpc_client)

    def get_screensaver_tasks(self) -> dict:
        return screensaver_tasks(client=self.rpc_client)

    # Versions
    def get_client_version(self) -> dict:
        return client_version(client=self.rpc_client)

    def get_client_update(self) -> dict:
        return client_update(client=self.rpc_client)

    def attach_project(self, name: str, url: str, key: str) -> dict:
        return attach_project(self.rpc_client, name, url, key)

    def poll_attach_project(self):
        return poll_attach_project(self.rpc_client)

    def update_project(self, url: str) -> dict:
        return update_project(self.rpc_client, url)

    def detach_project(self, url: str) -> dict:
        return detach_project(self.rpc_client, url)



def create_indexes(data, key: str, idx_key: str):
    data[key] = {d[idx_key]: d for d in data[key]}
    return data


def create_lists(data, keys: list[str]):
    for k in keys:
        if k not in data:
            data[k] = []
    return data


def flatten_data(data, key: str, subkey: str):
    if key in data and subkey in data[key]:
        data[key] = data[key][subkey] if data[key] else data[key]
    return data


def normalise_none_to_list(data, key: str):
    data[key] = [] if not data[key] else data[key]
    return data


def remove_key(data, key: str):
    del data[key]
    return data


def replace_none_string(data):
    for field in data:
        data[field] = data[field] if data[field] is not None else ""
    return data


def set_bools(data, keys: list[str]):
    for key in keys:
        data[key] = key in data
    return data



from marshmallow import Schema, fields, pre_load


class GenericResponse(Schema):
    success = fields.Bool()
    error = fields.Str(allow_none=True)

    @pre_load
    def _fill_success(self, data, **kwargs):
        data["success"] = "success" in data
        return data



from marshmallow import Schema, fields, pre_load

from boinc_client.models.helpers import flatten_data, normalise_none_to_list


class ActiveTask(Schema):
    active_task_state = fields.Int()
    app_version_num = fields.Int()
    slot = fields.Int()
    pid = fields.Int()
    scheduler_state = fields.Int()
    checkpoint_cpu_time = fields.Float()
    fraction_done = fields.Float()
    current_cpu_time = fields.Float()
    elapsed_time = fields.Float()
    swap_size = fields.Float()
    working_set_size = fields.Float()
    working_set_size_smoothed = fields.Float()
    page_fault_rate = fields.Float()
    bytes_sent = fields.Float()
    bytes_received = fields.Float()
    progress_rate = fields.Float()
    graphics_exec_path = fields.Str()
    slot_path = fields.Str()


class Result(Schema):
    name = fields.Str()
    wu_name = fields.Str()
    platform = fields.Str()
    version_num = fields.Int()
    plan_class = fields.Str(allow_none=True)
    project_url = fields.Url()
    final_cpu_time = fields.Float()
    final_elapsed_time = fields.Float()
    exit_status = fields.Int()
    state = fields.Int()
    report_deadline = fields.Float()
    received_time = fields.Float()
    estimated_cpu_time_remaining = fields.Float()
    edf_scheduled = fields.Bool()
    active_task = fields.Nested(ActiveTask())
    ready_to_report = fields.Bool()
    completed_time = fields.Float()

    @pre_load
    def _set_ready(self, data, **kwargs):
        data["ready_to_report"] = "ready_to_report" in data
        return data

    @pre_load
    def _set_scheduled(self, data, **kwargs):
        data["edf_scheduled"] = "edf_scheduled" in data
        return data


class Results(Schema):
    results = fields.Nested(Result(many=True))

    @pre_load
    def _a_normalise_none(self, data, **kwargs):
        return normalise_none_to_list(data, "results")

    @pre_load
    def _b_flatten_data(self, data, **kwargs):
        return flatten_data(data, "results", "result")



import xmltodict

from boinc_client.clients.rpc_client import RpcClient
from boinc_client.models.network_transfers import DailyTransfers
from boinc_client.models.projects_stats import ProjectStats


def daily_network_transfers(client: RpcClient) -> dict:
    """Show network traffic history of the BOINC client. Read from daily_xfer_history.xml."""
    rpc_resp = client.make_request("<get_daily_xfer_history/>")
    rpc_json = xmltodict.parse(rpc_resp, force_list="dx")
    return DailyTransfers().load(rpc_json)


def project_stats(client: RpcClient):
    """Get statistics for the projects the client is attached to."""
    rpc_resp = client.make_request("<get_statistics/>")
    rpc_json = xmltodict.parse(
        rpc_resp, force_list=("project_statistics", "daily_statistics")
    )
    return ProjectStats().load(rpc_json)



from marshmallow import Schema, fields


class ProjectAttachPollMessages(Schema):
    messages = fields.List(fields.Str(), allow_none=True, data_key="message")
    error_num = fields.Int()


class ProjectAttachPoll(Schema):
    project_attach_reply = fields.Nested(ProjectAttachPollMessages())



import re
import socket
from hashlib import md5

from lxml import etree

xml_parser = etree.XMLParser(resolve_entities=False)


class RpcClient:
    hostname: str
    port: int
    timeout: int
    password: str
    socket: socket.socket

    def __init__(
        self, hostname: str, port: int = 31416, timeout: int = 30, password: str = None
    ):
        self.hostname = hostname
        self.port = port
        self.timeout = timeout
        self.password = password
        self.socket = self.create_connection()

    def create_connection(self) -> socket:
        return socket.create_connection(
            address=(self.hostname, self.port), timeout=self.timeout
        )

    def authenticate(self) -> bool:
        if self.password:
            if self.send_password(nonce=self.get_nonce()) == "authorized":
                return True
        return False

    def get_nonce(self) -> str:
        # Docs: https://boinc.berkeley.edu/trac/wiki/GuiRpcProtocol#Authentication
        # Send the auth1 request and return the nonce
        auth1_resp = self._call(req_string="<auth1/>")
        nonce_resp = etree.fromstring(auth1_resp, parser=xml_parser)
        return nonce_resp[0].text

    def send_password(self, nonce: str) -> str:
        # Combine nonce with password and hash
        auth2_cred = md5(f"{nonce}{self.password}".encode()).hexdigest()
        auth2_req = f"<auth2>\n<nonce_hash>{auth2_cred}</nonce_hash>\n</auth2>"
        auth2_resp = etree.fromstring(
            self._call(req_string=auth2_req), parser=xml_parser
        )
        return auth2_resp[0].tag

    def make_request(self, req_str: str) -> str:
        response = self._call(req_string=req_str)
        return re.sub("<(/)?boinc_gui_rpc_reply>(\\n)?", "", response)

    def _call(self, req_string: str) -> str:
        # Send the request
        buf = f"<boinc_gui_rpc_request>\n{req_string}\n</boinc_gui_rpc_request>\003"
        try:
            self.socket.sendall(str.encode(buf))
        except (socket.error, socket.herror, socket.gaierror, socket.timeout):
            raise

        # Receive the response
        response = ""
        while True:
            try:
                resp = self.socket.recv(8192)
                if not resp:
                    raise socket.error("No data from socket")
                n = resp.find(b"\003")
                if not n == -1:
                    response += resp[:n].decode()
                    break
                else:
                    response += resp.decode()
            except socket.error:
                self.socket.close()
                raise
        return response



from datetime import datetime

from marshmallow import Schema, fields, post_load, pre_load

from boinc_client.models.helpers import flatten_data, normalise_none_to_list, remove_key


def _convert_epoch(epoch: str):
    epoch_date = datetime.fromtimestamp(float(epoch))
    return epoch_date.strftime("%Y-%m-%d")


class DailyStats(Schema):
    day = fields.Float()
    user_total_credit = fields.Float()
    user_expavg_credit = fields.Float()
    host_total_credit = fields.Float()
    host_expavg_credit = fields.Float()

    @post_load
    def _remove_day(self, data, **kwargs):
        return remove_key(data, "day")


class Project(Schema):
    master_url = fields.Str()
    daily_statistics = fields.Dict(fields.Str(), fields.Nested(DailyStats()))

    @pre_load
    def _create_dict_keys(self, data, **kwargs):
        data["daily_statistics"] = {
            _convert_epoch(day["day"]): day for day in data["daily_statistics"]
        }
        return data


class ProjectStats(Schema):
    project_stats = fields.Nested(Project(many=True), data_key="statistics")

    @pre_load
    def _a_normalise_none(self, data, **kwargs):
        return normalise_none_to_list(data, "statistics")

    @pre_load
    def _b_flatten_data(self, data, **kwargs):
        return flatten_data(data, "statistics", "project_statistics")
